@@ -55,6 +55,8 @@
 
 /*** Local data **************************************************************/
 
+static int ssl_verify_mode = SSL_VERIFY_PEER;
+
 /*
  * Error messages for resolver errors
  */
@@ -79,6 +81,12 @@ fetch_finderr(struct fetcherr *p, int e)
 	return (p);
 }
 
+void
+fetch_no_check_certificate(void)
+{
+	ssl_verify_mode = SSL_VERIFY_NONE;
+}
+
 /*
  * Set error code
  */
@@ -171,6 +179,30 @@ fetch_info(const char *fmt, ...)
 
 /*** Network-related utility functions ***************************************/
 
+uintmax_t
+fetch_parseuint(const char *str, const char **endptr, int radix, uintmax_t max)
+{
+	uintmax_t val = 0, maxx = max / radix, d;
+	const char *p;
+
+	for (p = str; isxdigit((unsigned char)*p); p++) {
+		unsigned char ch = (unsigned char)*p;
+		if (isdigit(ch))
+			d = ch - '0';
+		else	d = tolower(ch) - 'a' + 10;
+		if (d > radix || val > maxx) goto err;
+		val *= radix;
+		if (val > max-d) goto err;
+		val += d;
+	}
+	if (p == str || val > max) goto err;
+	*endptr = p;
+	return val;
+err:
+	*endptr = "\xff";
+	return 0;
+}
+
 /*
  * Return the default port for a scheme
  */
@@ -442,7 +474,7 @@ static int fetch_ssl_setup_peer_verification(SSL_CTX *ctx, int verbose)
 	else
 		SSL_CTX_set_default_verify_paths(ctx);
 
-	SSL_CTX_set_verify(ctx, SSL_VERIFY_PEER, 0);
+	SSL_CTX_set_verify(ctx, ssl_verify_mode, 0);
 	return 1;
 }
 
@@ -549,7 +581,8 @@ fetch_ssl(conn_t *conn, const struct url *URL, int verbose)
 				NULL) != 1) {
 			fprintf(stderr, "SSL certificate subject doesn't match host %s\n",
 				URL->host);
-			return -1;
+			if (ssl_verify_mode != SSL_VERIFY_NONE)
+				return -1;
 		}
 	}
 
