@@ -262,7 +262,7 @@ int apk_solver_commit_changeset(struct apk_database *db,
 	struct apk_change *change;
 	char buf[32];
 	const char *size_unit;
-	off_t humanized, size_diff = 0;
+	off_t humanized, size_diff = 0, download_size = 0;
 	int r, errors = 0;
 
 	assert(world);
@@ -279,12 +279,15 @@ int apk_solver_commit_changeset(struct apk_database *db,
 	memset(&prog, 0, sizeof(prog));
 	foreach_array_item(change, changeset->changes) {
 		count_change(change, &prog.total);
-		if (change->new_pkg)
+		if (change->new_pkg) {
 			size_diff += change->new_pkg->installed_size;
+			if (change->new_pkg != change->old_pkg &&
+			    !(change->new_pkg->repos & db->local_repos))
+				download_size += change->new_pkg->size;
+		}
 		if (change->old_pkg)
 			size_diff -= change->old_pkg->installed_size;
 	}
-	size_unit = apk_get_human_size(llabs(size_diff), &humanized);
 
 	if ((apk_verbosity > 1 || (apk_flags & APK_INTERACTIVE)) &&
 	    !(apk_flags & APK_SIMULATE)) {
@@ -299,6 +302,12 @@ int apk_solver_commit_changeset(struct apk_database *db,
 					   "The following packages will be upgraded");
 			r += dump_packages(changeset, cmp_reinstall,
 					   "The following packages will be reinstalled");
+			if (download_size) {
+				size_unit = apk_get_human_size(download_size, &humanized);
+				printf("Need to download %lld %s of packages.\n",
+					(long long)humanized, size_unit);
+			}
+			size_unit = apk_get_human_size(llabs(size_diff), &humanized);
 			printf("After this operation, %lld %s of %s.\n",
 				(long long)humanized,
 				size_unit,
@@ -310,7 +319,7 @@ int apk_solver_commit_changeset(struct apk_database *db,
 			printf("Do you want to continue [Y/n]? ");
 			fflush(stdout);
 			r = fgetc(stdin);
-			if (r != 'y' && r != 'Y' && r != '\n')
+			if (r != 'y' && r != 'Y' && r != '\n' && r != EOF)
 				return -1;
 		}
 	}
@@ -376,6 +385,8 @@ all_done:
 enum {
 	STATE_PRESENT		= 0x80000000,
 	STATE_MISSING		= 0x40000000,
+	STATE_VIRTUAL_ONLY	= 0x20000000,
+	STATE_INSTALLIF		= 0x10000000,
 	STATE_COUNT_MASK	= 0x0000ffff,
 };
 
@@ -421,7 +432,7 @@ static void print_pinning_errors(struct print_state *ps, struct apk_package *pkg
 	if (!(pkg->repos & db->available_repos)) {
 		label_start(ps, "masked in:");
 		apk_print_indented_fmt(&ps->i, "--no-network");
-	} else if (pkg->repos == BIT(APK_REPOSITORY_CACHED) && !(pkg->filename != NULL || pkg->installed_size == 0)) {
+	} else if (pkg->repos == BIT(APK_REPOSITORY_CACHED) && !pkg->filename) {
 		label_start(ps, "masked in:");
 		apk_print_indented_fmt(&ps->i, "cache");
 	} else {
@@ -511,7 +522,7 @@ static void analyze_package(struct print_state *ps, struct apk_package *pkg, uns
 		print_deps(ps, pkg, APK_DEP_SATISFIES);
 }
 
-static void analyze_name(struct print_state *ps, struct apk_name *name)
+static void analyze_missing_name(struct print_state *ps, struct apk_name *name)
 {
 	struct apk_name **pname0, *name0;
 	struct apk_provider *p0;
@@ -524,6 +535,10 @@ static void analyze_name(struct print_state *ps, struct apk_name *name)
 		snprintf(tmp, sizeof(tmp), "%s (virtual)", name->name);
 		ps->label = tmp;
 
+		label_start(ps, "note:");
+		apk_print_indented_words(&ps->i, "please select one of the 'provided by' packages explicitly");
+		label_end(ps);
+
 		label_start(ps, "provided by:");
 		foreach_array_item(p0, name->providers)
 			p0->pkg->name->state_int++;
@@ -585,12 +600,91 @@ static void analyze_deps(struct print_state *ps, struct apk_dependency_array *de
 
 	foreach_array_item(d0, deps) {
 		name0 = d0->name;
-		if (d0->conflict)
-			continue;
+		if (d0->conflict) continue;
 		if ((name0->state_int & (STATE_PRESENT | STATE_MISSING)) != 0)
 			continue;
 		name0->state_int |= STATE_MISSING;
-		analyze_name(ps, name0);
+		analyze_missing_name(ps, name0);
+	}
+}
+
+static void discover_deps(struct apk_dependency_array *deps);
+static void discover_name(struct apk_name *name, int pkg_state);
+
+static void discover_reverse_iif(struct apk_name *name)
+{
+	struct apk_name **pname0, *name0;
+	struct apk_dependency *d;
+	struct apk_provider *p;
+
+	foreach_array_item(pname0, name->rinstall_if) {
+		name0 = *pname0;
+
+		foreach_array_item(p, name0->providers) {
+			int ok = 1;
+			if (!p->pkg->marked) continue;
+			if (p->pkg->install_if->num == 0) continue;
+			foreach_array_item(d, p->pkg->install_if) {
+				if (!!d->conflict == !!(d->name->state_int & STATE_PRESENT)) {
+					ok = 0;
+					break;
+				}
+			}
+			if (ok) {
+				discover_name(p->pkg->name, STATE_INSTALLIF);
+				foreach_array_item(d, p->pkg->provides)
+					discover_name(d->name, STATE_INSTALLIF);
+			}
+		}
+	}
+}
+
+static int is_name_concrete(struct apk_package *pkg, struct apk_name *name)
+{
+	struct apk_dependency *d;
+	if (pkg->name == name) return 1;
+	foreach_array_item(d, pkg->provides) {
+		if (d->name != name) continue;
+		if (d->version == &apk_atom_null) continue;
+		return 1;
+	}
+	return 0;
+}
+
+static void discover_name(struct apk_name *name, int pkg_state)
+{
+	struct apk_provider *p;
+	struct apk_dependency *d;
+
+	foreach_array_item(p, name->providers) {
+		int state = pkg_state;
+		if (!p->pkg->marked) continue;
+		if (state == STATE_PRESENT && !p->pkg->provider_priority &&
+		    !is_name_concrete(p->pkg, name))
+			state = STATE_VIRTUAL_ONLY;
+		if (p->pkg->state_int & state) continue;
+		p->pkg->state_int |= state;
+
+		p->pkg->name->state_int |= state;
+		foreach_array_item(d, p->pkg->provides)
+			d->name->state_int |= state;
+
+		discover_deps(p->pkg->depends);
+		if (state == STATE_PRESENT || state == STATE_INSTALLIF) {
+			discover_reverse_iif(p->pkg->name);
+			foreach_array_item(d, p->pkg->provides)
+				discover_reverse_iif(d->name);
+		}
+	}
+}
+
+static void discover_deps(struct apk_dependency_array *deps)
+{
+	struct apk_dependency *d;
+
+	foreach_array_item(d, deps) {
+		if (d->conflict) continue;
+		discover_name(d->name, STATE_PRESENT);
 	}
 }
 
@@ -600,7 +694,6 @@ void apk_solver_print_errors(struct apk_database *db,
 {
 	struct print_state ps;
 	struct apk_change *change;
-	struct apk_dependency *p;
 
 	/* ERROR: unsatisfiable dependencies:
 	 *   name:
@@ -644,13 +737,9 @@ void apk_solver_print_errors(struct apk_database *db,
 	/* Construct information about names */
 	foreach_array_item(change, changeset->changes) {
 		struct apk_package *pkg = change->new_pkg;
-		if (pkg == NULL)
-			continue;
-		pkg->marked = 1;
-		pkg->name->state_int |= STATE_PRESENT;
-		foreach_array_item(p, pkg->provides)
-			p->name->state_int |= STATE_PRESENT;
+		if (pkg) pkg->marked = 1;
 	}
+	discover_deps(world);
 
 	/* Analyze is package, and missing names referred to */
 	ps = (struct print_state) {
@@ -660,8 +749,7 @@ void apk_solver_print_errors(struct apk_database *db,
 	analyze_deps(&ps, world);
 	foreach_array_item(change, changeset->changes) {
 		struct apk_package *pkg = change->new_pkg;
-		if (pkg == NULL)
-			continue;
+		if (!pkg) continue;
 		analyze_package(&ps, pkg, change->new_repository_tag);
 		analyze_deps(&ps, pkg->depends);
 	}
