@@ -14,7 +14,6 @@
 #include <libgen.h>
 #include <limits.h>
 #include <unistd.h>
-#include <malloc.h>
 #include <string.h>
 #include <stdlib.h>
 #include <signal.h>
@@ -484,29 +483,29 @@ static struct apk_db_file *apk_db_file_get(struct apk_database *db,
 	return file;
 }
 
+static void add_name_to_array(struct apk_name *name, struct apk_name_array **a)
+{
+	struct apk_name **n;
+
+	foreach_array_item(n, *a)
+		if (*n == name) return;
+	*apk_name_array_add(a) = name;
+}
+
 static void apk_db_pkg_rdepends(struct apk_database *db, struct apk_package *pkg)
 {
-	struct apk_name *rname, **rd;
+	struct apk_name *rname;
 	struct apk_dependency *d;
 
 	foreach_array_item(d, pkg->depends) {
 		rname = d->name;
 		rname->is_dependency |= !d->conflict;
-		foreach_array_item(rd, rname->rdepends)
-			if (*rd == pkg->name)
-				goto rdeps_done;
-		*apk_name_array_add(&rname->rdepends) = pkg->name;
-rdeps_done: ;
+		add_name_to_array(pkg->name, &rname->rdepends);
 	}
 	foreach_array_item(d, pkg->install_if) {
 		rname = d->name;
-		foreach_array_item(rd, rname->rinstall_if)
-			if (*rd == pkg->name)
-				goto riif_done;
-		*apk_name_array_add(&rname->rinstall_if) = pkg->name;
-riif_done: ;
+		add_name_to_array(pkg->name, &rname->rinstall_if);
 	}
-	return;
 }
 
 static inline void add_provider(struct apk_name *name, struct apk_provider p)
@@ -523,10 +522,8 @@ struct apk_package *apk_db_pkg_add(struct apk_database *db, struct apk_package *
 
 	if (!pkg->license) pkg->license = &apk_atom_null;
 
-	/* Set as "cached" if installing from specified file, and
-	 * for virtual packages */
-	if (pkg->filename != NULL || pkg->installed_size == 0)
-		pkg->repos |= BIT(APK_REPOSITORY_CACHED);
+	// Set as "cached" if installing from specified file
+	if (pkg->filename) pkg->repos |= BIT(APK_REPOSITORY_CACHED);
 
 	idb = apk_hash_get(&db->available.packages, APK_BLOB_CSUM(pkg->csum));
 	if (idb == NULL) {
@@ -933,7 +930,7 @@ static int apk_db_write_fdb(struct apk_database *db, struct apk_ostream *os)
 	struct apk_db_dir_instance *diri;
 	struct apk_db_file *file;
 	struct hlist_node *c1, *c2;
-	char buf[1024];
+	char buf[1024+PATH_MAX];
 	apk_blob_t bbuf = APK_BLOB_BUF(buf);
 	int r;
 
@@ -976,6 +973,12 @@ static int apk_db_write_fdb(struct apk_database *db, struct apk_ostream *os)
 			if (diri->acl != apk_default_acl_dir)
 				apk_blob_push_db_acl(&bbuf, 'M', diri->acl);
 
+			bbuf = apk_blob_pushed(APK_BLOB_BUF(buf), bbuf);
+			if (APK_BLOB_IS_NULL(bbuf)) return -ENOBUFS;
+			r = apk_ostream_write(os, bbuf.ptr, bbuf.len);
+			if (r != bbuf.len) return r < 0 ? r : -ENOSPC;
+			bbuf = APK_BLOB_BUF(buf);
+
 			hlist_for_each_entry(file, c2, &diri->owned_files, diri_files_list) {
 				apk_blob_push_blob(&bbuf, APK_BLOB_STR("R:"));
 				apk_blob_push_blob(&bbuf, APK_BLOB_PTR_LEN(file->name, file->namelen));
@@ -990,13 +993,12 @@ static int apk_db_write_fdb(struct apk_database *db, struct apk_ostream *os)
 					apk_blob_push_blob(&bbuf, APK_BLOB_STR("\n"));
 				}
 
-				if (apk_ostream_write(os, buf, bbuf.ptr - buf) != bbuf.ptr - buf)
-					return -EIO;
+				bbuf = apk_blob_pushed(APK_BLOB_BUF(buf), bbuf);
+				if (APK_BLOB_IS_NULL(bbuf)) return -ENOBUFS;
+				r = apk_ostream_write(os, bbuf.ptr, bbuf.len);
+				if (r != bbuf.len) return r < 0 ? r : -ENOSPC;
 				bbuf = APK_BLOB_BUF(buf);
 			}
-			if (apk_ostream_write(os, buf, bbuf.ptr - buf) != bbuf.ptr - buf)
-				return -EIO;
-			bbuf = APK_BLOB_BUF(buf);
 		}
 		apk_ostream_write(os, "\n", 1);
 	}
@@ -1012,7 +1014,6 @@ static int apk_db_scriptdb_write(struct apk_database *db, struct apk_ostream *os
 	char filename[256];
 	apk_blob_t bfn;
 	int r, i;
-	time_t now = time(NULL);
 
 	list_for_each_entry(ipkg, &db->installed.packages, installed_pkgs_list) {
 		pkg = ipkg->pkg;
@@ -1025,7 +1026,6 @@ static int apk_db_scriptdb_write(struct apk_database *db, struct apk_ostream *os
 				.name = filename,
 				.size = ipkg->script[i].len,
 				.mode = 0755 | S_IFREG,
-				.mtime = now,
 			};
 			/* The scripts db expects file names in format:
 			 * pkg-version.<hexdump of package checksum>.action */
@@ -1217,8 +1217,7 @@ static int apk_db_index_write_nr_cache(struct apk_database *db)
 	struct apk_ostream *os;
 	int r;
 
-	if (!apk_db_cache_active(db))
-		return 0;
+	if (!apk_db_cache_active(db)) return 0;
 
 	/* Write list of installed non-repository packages to
 	 * cached index file */
@@ -1231,16 +1230,14 @@ static int apk_db_index_write_nr_cache(struct apk_database *db)
 	ctx.os = os;
 	list_for_each_entry(ipkg, &db->installed.packages, installed_pkgs_list) {
 		struct apk_package *pkg = ipkg->pkg;
-		if (pkg->repos != BIT(APK_REPOSITORY_CACHED))
-			continue;
-		r = write_index_entry(pkg, &ctx);
-		if (r != 0)
-			return r;
+		if ((pkg->repos == BIT(APK_REPOSITORY_CACHED) ||
+		     (pkg->repos == 0 && !pkg->installed_size))) {
+			r = write_index_entry(pkg, &ctx);
+			if (r != 0) return r;
+		}
 	}
 	r = apk_ostream_close(os);
-	if (r < 0)
-		return r;
-
+	if (r < 0) return r;
 	return ctx.count;
 }
 
@@ -1670,7 +1667,10 @@ int apk_db_open(struct apk_database *db, struct apk_db_options *dbopts)
 			mkdirat(db->root_fd, "var/cache", 0755);
 			mkdirat(db->root_fd, "var/cache/apk", 0755);
 			db->cache_fd = openat(db->root_fd, db->cache_dir, O_RDONLY | O_CLOEXEC);
-			if (db->cache_fd < 0) goto ret_errno;
+			if (db->cache_fd < 0) {
+				if (dbopts->open_flags & APK_OPENF_WRITE) goto ret_errno;
+				db->cache_fd = -EAPKCACHE;
+			}
 		}
 	}
 
@@ -1715,7 +1715,7 @@ int apk_db_open(struct apk_database *db, struct apk_db_options *dbopts)
 			add_repos_from_file(db, AT_FDCWD, dbopts->repositories_file);
 		}
 
-		if (db->repo_update_counter)
+		if (db->repositories.updated > 0)
 			apk_db_index_write_nr_cache(db);
 
 		apk_hash_foreach(&db->available.names, apk_db_name_rdepends, db);
@@ -1854,16 +1854,11 @@ void apk_db_close(struct apk_database *db)
 		db->cache_remount_dir = NULL;
 	}
 
-	if (db->keys_fd)
-		close(db->keys_fd);
-	if (db->cache_fd)
-		close(db->cache_fd);
-	if (db->root_fd)
-		close(db->root_fd);
-	if (db->lock_fd)
-		close(db->lock_fd);
-	if (db->root != NULL)
-		free(db->root);
+	if (db->keys_fd > 0) close(db->keys_fd);
+	if (db->cache_fd > 0) close(db->cache_fd);
+	if (db->root_fd > 0) close(db->root_fd);
+	if (db->lock_fd > 0) close(db->lock_fd);
+	free(db->root);
 }
 
 int apk_db_get_tag_id(struct apk_database *db, apk_blob_t tag)
@@ -1980,9 +1975,15 @@ int apk_db_run_script(struct apk_database *db, char *fn, char **argv)
 	return 0;
 }
 
-static int update_permissions(apk_hash_item item, void *ctx)
+struct update_permissions_ctx {
+	struct apk_database *db;
+	unsigned int errors;
+};
+
+static int update_permissions(apk_hash_item item, void *pctx)
 {
-	struct apk_database *db = (struct apk_database *) ctx;
+	struct update_permissions_ctx *ctx = pctx;
+	struct apk_database *db = ctx->db;
 	struct apk_db_dir *dir = (struct apk_db_dir *) item;
 	struct stat st;
 	int r;
@@ -1993,9 +1994,11 @@ static int update_permissions(apk_hash_item item, void *ctx)
 
 	r = fstatat(db->root_fd, dir->name, &st, AT_SYMLINK_NOFOLLOW);
 	if (r < 0 || (st.st_mode & 07777) != (dir->mode & 07777))
-		fchmodat(db->root_fd, dir->name, dir->mode, 0);
+		if (fchmodat(db->root_fd, dir->name, dir->mode, 0) < 0)
+			ctx->errors++;
 	if (r < 0 || st.st_uid != dir->uid || st.st_gid != dir->gid)
-		fchownat(db->root_fd, dir->name, dir->uid, dir->gid, 0);
+		if (fchownat(db->root_fd, dir->name, dir->uid, dir->gid, 0) < 0)
+			ctx->errors++;
 
 	return 0;
 }
@@ -2006,6 +2009,9 @@ void apk_db_update_directory_permissions(struct apk_database *db)
 	struct apk_db_dir_instance *diri;
 	struct apk_db_dir *dir;
 	struct hlist_node *dc, *dn;
+	struct update_permissions_ctx ctx = {
+		.db = db,
+	};
 
 	list_for_each_entry(ipkg, &db->installed.packages, installed_pkgs_list) {
 		hlist_for_each_entry_safe(diri, dc, dn, &ipkg->owned_dirs, pkg_dirs_list) {
@@ -2020,7 +2026,8 @@ void apk_db_update_directory_permissions(struct apk_database *db)
 			apk_db_dir_apply_diri_permissions(diri);
 		}
 	}
-	apk_hash_foreach(&db->installed.dirs, update_permissions, db);
+	apk_hash_foreach(&db->installed.dirs, update_permissions, &ctx);
+	if (ctx.errors) apk_error("%d errors updating directory permissions", ctx.errors);
 }
 
 int apk_db_cache_active(struct apk_database *db)
@@ -2163,24 +2170,6 @@ struct apk_repository *apk_db_select_repo(struct apk_database *db,
 	return &db->repos[APK_REPOSITORY_CACHED];
 }
 
-static int apk_repository_update(struct apk_database *db, struct apk_repository *repo)
-{
-	struct apk_url_print urlp;
-	int r, verify = (apk_flags & APK_ALLOW_UNTRUSTED) ? APK_SIGN_NONE : APK_SIGN_VERIFY;
-
-	r = apk_cache_download(db, repo, NULL, verify, 1, NULL, NULL);
-	if (r == -EALREADY) return 0;
-	if (r != 0) {
-		apk_url_parse(&urlp, repo->url);
-		apk_error(URL_FMT ": %s", URL_PRINTF(urlp), apk_error_str(r));
-		db->repo_update_errors++;
-	} else {
-		db->repo_update_counter++;
-	}
-
-	return r;
-}
-
 struct apkindex_ctx {
 	struct apk_database *db;
 	struct apk_sign_ctx sctx;
@@ -2252,8 +2241,10 @@ int apk_db_add_repository(apk_database_t _db, apk_blob_t _repository)
 	struct apk_repository *repo;
 	struct apk_url_print urlp;
 	apk_blob_t brepo, btag;
-	int repo_num, r, targz = 1, tag_id = 0;
+	int repo_num, r, tag_id = 0, atfd = AT_FDCWD, update_error = 0;
 	char buf[PATH_MAX], *url;
+	const int verify = (apk_flags & APK_ALLOW_UNTRUSTED) ? APK_SIGN_NONE : APK_SIGN_VERIFY;
+	const char *error_action = "opening";
 
 	brepo = _repository;
 	btag = APK_BLOB_NULL;
@@ -2289,15 +2280,29 @@ int apk_db_add_repository(apk_database_t _db, apk_blob_t _repository)
 
 	apk_blob_checksum(brepo, apk_checksum_default(), &repo->csum);
 
-	if (apk_url_local_file(repo->url) == NULL) {
+	int is_remote = (apk_url_local_file(repo->url) == NULL);
+	if (is_remote) {
 		if (!(apk_flags & APK_NO_NETWORK))
 			db->available_repos |= BIT(repo_num);
 		if (apk_flags & APK_NO_CACHE) {
+			error_action = "fetching";
 			r = apk_repo_format_real_url(db->arch, repo, NULL, buf, sizeof(buf), &urlp);
 			if (r == 0) apk_message("fetch " URL_FMT, URL_PRINTF(urlp));
 		} else {
-			if (db->autoupdate) apk_repository_update(db, repo);
+			error_action = "opening from cache";
+			if (db->autoupdate) {
+				update_error = apk_cache_download(db, repo, NULL, verify, 1, NULL, NULL);
+				switch (update_error) {
+				case 0:
+					db->repositories.updated++;
+					break;
+				case -EALREADY:
+					update_error = 0;
+					break;
+				}
+			}
 			r = apk_repo_format_cache_index(APK_BLOB_BUF(buf), repo);
+			atfd = db->cache_fd;
 		}
 	} else {
 		db->local_repos |= BIT(repo_num);
@@ -2305,14 +2310,26 @@ int apk_db_add_repository(apk_database_t _db, apk_blob_t _repository)
 		r = apk_repo_format_real_url(db->arch, repo, NULL, buf, sizeof(buf), &urlp);
 	}
 	if (r == 0) {
-		r = load_index(db, apk_istream_from_fd_url(db->cache_fd, buf), targz, repo_num);
+		r = load_index(db, apk_istream_from_fd_url(atfd, buf), 1, repo_num);
+	}
+
+	if (r || update_error) {
+		if (is_remote) {
+			if (r) db->repositories.unavailable++;
+			else db->repositories.stale++;
+		}
+		apk_url_parse(&urlp, repo->url);
+		if (update_error)
+			error_action = r ? "updating and opening" : "updating";
+		else
+			update_error = r;
+		apk_warning("%s " URL_FMT ": %s", error_action, URL_PRINTF(urlp),
+			apk_error_str(update_error));
 	}
 
 	if (r != 0) {
-		apk_url_parse(&urlp, repo->url);
-		apk_warning("Ignoring " URL_FMT ": %s", URL_PRINTF(urlp), apk_error_str(r));
 		db->available_repos &= ~BIT(repo_num);
-		r = 0;
+		if (is_remote) db->repositories.unavailable++;
 	} else {
 		db->repo_tags[tag_id].allowed_repos |= BIT(repo_num);
 	}
@@ -2448,8 +2465,10 @@ static int apk_db_install_archive_entry(void *_ctx,
 		if (ae->name[0] != '.') return 0;
 		if (strcmp(ae->name, ".PKGINFO") == 0) {
 			apk_blob_t l, token = APK_BLOB_STR("\n");
-			while (!APK_BLOB_IS_NULL(l = apk_istream_get_delim(is, token)))
-				read_info_line(ctx, l);
+			while (!APK_BLOB_IS_NULL(l = apk_istream_get_delim(is, token))) {
+				r = read_info_line(ctx, l);
+				if (r < 0) return r;
+			}
 			return 0;
 		}
 		r = apk_script_type(&ae->name[1]);
@@ -2672,7 +2691,8 @@ static void apk_db_purge_pkg(struct apk_database *db,
 				.filename = APK_BLOB_PTR_LEN(file->name, file->namelen),
 			};
 			hash = apk_blob_hash_seed(key.filename, diri->dir->hash);
-			if ((diri->dir->protect_mode == APK_PROTECT_NONE) ||
+			if (!is_installed ||
+			    (diri->dir->protect_mode == APK_PROTECT_NONE) ||
 			    (apk_flags & APK_PURGE) ||
 			    (file->csum.type != APK_CHECKSUM_NONE &&
 			     apk_fileinfo_get(db->root_fd, name, APK_FI_NOFOLLOW | file->csum.type, &fi, &db->atoms) == 0 &&
