@@ -12,7 +12,6 @@
 #include <fcntl.h>
 #include <endian.h>
 #include <unistd.h>
-#include <malloc.h>
 #include <dirent.h>
 #include <stdint.h>
 #include <sys/mman.h>
@@ -21,6 +20,7 @@
 #include <sys/xattr.h>
 #include <pwd.h>
 #include <grp.h>
+#include <limits.h>
 
 #include "apk_defines.h"
 #include "apk_io.h"
@@ -34,6 +34,11 @@
 
 size_t apk_io_bufsize = 128*1024;
 
+static inline int atfd_error(int atfd)
+{
+	return atfd < -1 && atfd != AT_FDCWD;
+}
+
 static void apk_file_meta_from_fd(int fd, struct apk_file_meta *meta)
 {
 	struct stat st;
@@ -314,6 +319,7 @@ struct apk_istream *apk_istream_tee(struct apk_istream *from, int atfd, const ch
 	int fd, r;
 
 	if (IS_ERR_OR_NULL(from)) return ERR_CAST(from);
+	if (atfd_error(atfd)) return ERR_PTR(atfd);
 
 	fd = openat(atfd, to, O_CREAT | O_RDWR | O_TRUNC | O_CLOEXEC,
 		    S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
@@ -482,6 +488,8 @@ struct apk_istream *apk_istream_from_file(int atfd, const char *file)
 {
 	int fd;
 
+	if (atfd_error(atfd)) return ERR_PTR(atfd);
+
 	fd = openat(atfd, file, O_RDONLY | O_CLOEXEC);
 	if (fd < 0) return ERR_PTR(-errno);
 
@@ -604,6 +612,8 @@ apk_blob_t apk_blob_from_file(int atfd, const char *file)
 	struct stat st;
 	char *buf;
 
+	if (atfd_error(atfd)) return APK_BLOB_NULL;
+
 	fd = openat(atfd, file, O_RDONLY | O_CLOEXEC);
 	if (fd < 0)
 		return APK_BLOB_NULL;
@@ -631,6 +641,8 @@ int apk_blob_to_file(int atfd, const char *file, apk_blob_t b, unsigned int flag
 {
 	int fd, r, len;
 
+	if (atfd_error(atfd)) return atfd;
+
 	fd = openat(atfd, file, O_CREAT | O_WRONLY | O_CLOEXEC, 0644);
 	if (fd < 0)
 		return -errno;
@@ -707,6 +719,9 @@ int apk_fileinfo_get(int atfd, const char *filename, unsigned int flags,
 	unsigned int xattr_checksum = (flags >> 8) & 0xff;
 	int atflags = 0;
 
+	if (atfd_error(atfd)) return atfd;
+
+	memset(fi, 0, sizeof *fi);
 	if (flags & APK_FI_NOFOLLOW)
 		atflags |= AT_SYMLINK_NOFOLLOW;
 
@@ -808,8 +823,10 @@ int apk_dir_foreach_file(int dirfd, apk_dir_file_cb cb, void *ctx)
 		return -1;
 
 	dir = fdopendir(dirfd);
-	if (dir == NULL)
+	if (!dir) {
+		close(dirfd);
 		return -1;
+	}
 
 	/* We get called here with dup():ed fd. Since they all refer to
 	 * same object, we need to rewind so subsequent calls work. */
@@ -957,6 +974,8 @@ struct apk_ostream *apk_ostream_to_file(int atfd,
 	struct apk_ostream *os;
 	int fd;
 
+	if (atfd_error(atfd)) return ERR_PTR(atfd);
+
 	fd = openat(atfd, tmpfile ?: file, O_CREAT | O_RDWR | O_TRUNC | O_CLOEXEC, mode);
 	if (fd < 0) return ERR_PTR(-errno);
 
@@ -1095,7 +1114,20 @@ void apk_id_cache_reset(struct apk_id_cache *idc)
 		idc->genid = 1;
 }
 
-uid_t apk_resolve_uid(struct apk_id_cache *idc, const char *username, uid_t default_uid)
+static FILE *fopenat(int dirfd, const char *pathname)
+{
+	FILE *f;
+	int fd;
+
+	fd = openat(dirfd, pathname, O_RDONLY|O_CLOEXEC);
+	if (fd < 0) return NULL;
+
+	f = fdopen(fd, "r");
+	if (!f) close(fd);
+	return f;
+}
+
+uid_t apk_resolve_uid(struct apk_id_cache *idc, apk_blob_t username, uid_t default_uid)
 {
 #ifdef HAVE_FGETPWENT_R
 	char buf[1024];
@@ -1105,7 +1137,7 @@ uid_t apk_resolve_uid(struct apk_id_cache *idc, const char *username, uid_t defa
 	struct passwd *pwd;
 	FILE *in;
 
-	ci = resolve_cache_item(&idc->uid_cache, APK_BLOB_STR(username));
+	ci = resolve_cache_item(&idc->uid_cache, username);
 	if (ci == NULL)
 		return default_uid;
 
@@ -1113,8 +1145,8 @@ uid_t apk_resolve_uid(struct apk_id_cache *idc, const char *username, uid_t defa
 		ci->genid = idc->genid;
 		ci->uid = -1;
 
-		in = fdopen(openat(idc->root_fd, "etc/passwd", O_RDONLY|O_CLOEXEC), "r");
-		if (in != NULL) {
+		in = fopenat(idc->root_fd, "etc/passwd");
+		if (in) {
 			do {
 #ifdef HAVE_FGETPWENT_R
 				fgetpwent_r(in, &pwent, buf, sizeof(buf), &pwd);
@@ -1123,7 +1155,7 @@ uid_t apk_resolve_uid(struct apk_id_cache *idc, const char *username, uid_t defa
 #endif
 				if (pwd == NULL)
 					break;
-				if (strcmp(pwd->pw_name, username) == 0) {
+				if (apk_blob_compare(APK_BLOB_STR(pwd->pw_name), username) == 0) {
 					ci->uid = pwd->pw_uid;
 					break;
 				}
@@ -1138,7 +1170,7 @@ uid_t apk_resolve_uid(struct apk_id_cache *idc, const char *username, uid_t defa
 	return default_uid;
 }
 
-uid_t apk_resolve_gid(struct apk_id_cache *idc, const char *groupname, uid_t default_gid)
+uid_t apk_resolve_gid(struct apk_id_cache *idc, apk_blob_t groupname, uid_t default_gid)
 {
 #ifdef HAVE_FGETGRENT_R
 	char buf[1024];
@@ -1148,7 +1180,7 @@ uid_t apk_resolve_gid(struct apk_id_cache *idc, const char *groupname, uid_t def
 	struct group *grp;
 	FILE *in;
 
-	ci = resolve_cache_item(&idc->gid_cache, APK_BLOB_STR(groupname));
+	ci = resolve_cache_item(&idc->gid_cache, groupname);
 	if (ci == NULL)
 		return default_gid;
 
@@ -1156,8 +1188,8 @@ uid_t apk_resolve_gid(struct apk_id_cache *idc, const char *groupname, uid_t def
 		ci->genid = idc->genid;
 		ci->gid = -1;
 
-		in = fdopen(openat(idc->root_fd, "etc/group", O_RDONLY|O_CLOEXEC), "r");
-		if (in != NULL) {
+		in = fopenat(idc->root_fd, "etc/group");
+		if (in) {
 			do {
 #ifdef HAVE_FGETGRENT_R
 				fgetgrent_r(in, &grent, buf, sizeof(buf), &grp);
@@ -1166,7 +1198,7 @@ uid_t apk_resolve_gid(struct apk_id_cache *idc, const char *groupname, uid_t def
 #endif
 				if (grp == NULL)
 					break;
-				if (strcmp(grp->gr_name, groupname) == 0) {
+				if (apk_blob_compare(APK_BLOB_STR(grp->gr_name), groupname) == 0) {
 					ci->gid = grp->gr_gid;
 					break;
 				}
