@@ -12,7 +12,6 @@
 #include <ctype.h>
 #include <stdio.h>
 #include <limits.h>
-#include <malloc.h>
 #include <string.h>
 #include <stdlib.h>
 #include <unistd.h>
@@ -840,7 +839,7 @@ int apk_pkg_add_info(struct apk_database *db, struct apk_package *pkg,
 		return 2;
 	}
 	if (APK_BLOB_IS_NULL(value))
-		return -1;
+		return -EAPKFORMAT;
 	return 0;
 }
 
@@ -876,14 +875,11 @@ static int read_info_line(void *ctx, apk_blob_t line)
 	if (!apk_blob_split(line, APK_BLOB_STR(" = "), &l, &r))
 		return 0;
 
-	for (i = 0; i < ARRAY_SIZE(fields); i++) {
-		if (apk_blob_compare(APK_BLOB_STR(fields[i].str), l) == 0) {
-			apk_pkg_add_info(ri->db, ri->pkg, fields[i].field, r);
-			return 0;
-		}
-	}
-	apk_sign_ctx_parse_pkginfo_line(ri->sctx, line);
+	for (i = 0; i < ARRAY_SIZE(fields); i++)
+		if (apk_blob_compare(APK_BLOB_STR(fields[i].str), l) == 0)
+			return apk_pkg_add_info(ri->db, ri->pkg, fields[i].field, r);
 
+	apk_sign_ctx_parse_pkginfo_line(ri->sctx, line);
 	return 0;
 }
 
@@ -904,11 +900,13 @@ static int read_info_entry(void *ctx, const struct apk_file_info *ae,
 	if (strcmp(ae->name, ".PKGINFO") == 0) {
 		/* APK 2.0 format */
 		apk_blob_t l, token = APK_BLOB_STR("\n");
-		while (!APK_BLOB_IS_NULL(l = apk_istream_get_delim(is, token)))
-			read_info_line(ctx, l);
+		while (!APK_BLOB_IS_NULL(l = apk_istream_get_delim(is, token))) {
+			r = read_info_line(ctx, l);
+			if (r < 0) return r;
+		}
 	} else if (strcmp(ae->name, ".INSTALL") == 0) {
-		apk_warning("Package '%s-%s' contains deprecated .INSTALL",
-				pkg->name->name, pkg->version);
+		apk_warning("Package '%s-" BLOB_FMT "' contains deprecated .INSTALL",
+				pkg->name->name, BLOB_PRINTF(*pkg->version));
 	}
 
 	return 0;
@@ -999,10 +997,33 @@ int apk_ipkg_add_script(struct apk_installed_package *ipkg,
 	return 0;
 }
 
+static inline int make_dirs(int root_fd, const char *dirname, mode_t dirmode, mode_t parentmode)
+{
+	char parentdir[PATH_MAX], *slash;
+
+	if (faccessat(root_fd, dirname, F_OK, 0) == 0) return 0;
+	if (mkdirat(root_fd, dirname, dirmode) == 0) return 0;
+	if (errno != ENOENT || !parentmode) return -1;
+
+	slash = strrchr(dirname, '/');
+	if (!slash || slash == dirname || slash-dirname+1 >= sizeof parentdir) return -1;
+	strlcpy(parentdir, dirname, slash-dirname+1);
+	if (make_dirs(root_fd, parentdir, parentmode, parentmode) < 0) return -1;
+	return mkdirat(root_fd, dirname, dirmode);
+}
+
 void apk_ipkg_run_script(struct apk_installed_package *ipkg,
 			 struct apk_database *db,
 			 unsigned int type, char **argv)
 {
+	// script_exec_dir is the directory to which the script is extracted,
+	// executed from, and removed. It needs to not be 'noexec' mounted, and
+	// preferably a tmpfs disk, or something that could be wiped in boot.
+	// Originally this was /tmp, but it is often suggested to be 'noexec'.
+	// Then changed ro /var/cache/misc, but that is also often 'noexec'.
+	// /run was consider as it's tmpfs, but it also might be changing to 'noexec'.
+	// So use for now /lib/apk/exec even if it is not of temporary nature.
+	static const char script_exec_dir[] = "lib/apk/exec";
 	struct apk_package *pkg = ipkg->pkg;
 	char fn[PATH_MAX];
 	int fd, root_fd = db->root_fd;
@@ -1012,18 +1033,17 @@ void apk_ipkg_run_script(struct apk_installed_package *ipkg,
 
 	argv[0] = (char *) apk_script_types[type];
 
-	/* Avoid /tmp as it can be mounted noexec */
-	snprintf(fn, sizeof(fn), "var/cache/misc/" PKG_VER_FMT ".%s",
-		PKG_VER_PRINTF(pkg),
+	snprintf(fn, sizeof(fn), "%s/" PKG_VER_FMT ".%s",
+		script_exec_dir, PKG_VER_PRINTF(pkg),
 		apk_script_types[type]);
 
 	if ((apk_flags & (APK_NO_SCRIPTS | APK_SIMULATE)) != 0)
 		return;
 
-	apk_message("Executing %s", &fn[15]);
+	apk_message("Executing %s", &fn[strlen(script_exec_dir)+1]);
 	fd = openat(root_fd, fn, O_CREAT|O_RDWR|O_TRUNC|O_CLOEXEC, 0755);
 	if (fd < 0) {
-		mkdirat(root_fd, "var/cache/misc", 0755);
+		make_dirs(root_fd, script_exec_dir, 0700, 0755);
 		fd = openat(root_fd, fn, O_CREAT|O_RDWR|O_TRUNC|O_CLOEXEC, 0755);
 		if (fd < 0) goto err_log;
 	}
@@ -1097,7 +1117,7 @@ static int write_depends(struct apk_ostream *os, const char *field,
 int apk_pkg_write_index_entry(struct apk_package *info,
 			      struct apk_ostream *os)
 {
-	char buf[512];
+	char buf[2048];
 	apk_blob_t bbuf = APK_BLOB_BUF(buf);
 
 	apk_blob_push_blob(&bbuf, APK_BLOB_STR("C:"));
