@@ -51,6 +51,7 @@ void apk_solver_set_name_flags(struct apk_name *name,
 {
 	struct apk_provider *p;
 
+	name->solver_flags_set = 1;
 	foreach_array_item(p, name->providers) {
 		struct apk_package *pkg = p->pkg;
 		dbg_printf("marking '" PKG_VER_FMT "' = 0x%04x / 0x%04x\n",
@@ -118,8 +119,7 @@ static void reevaluate_reverse_deps(struct apk_solver_state *ss, struct apk_name
 
 	foreach_array_item(pname0, name->rdepends) {
 		name0 = *pname0;
-		if (!name0->ss.seen)
-			continue;
+		if (!name0->ss.seen) continue;
 		name0->ss.reevaluate_deps = 1;
 		queue_dirty(ss, name0);
 	}
@@ -131,15 +131,21 @@ static void reevaluate_reverse_installif(struct apk_solver_state *ss, struct apk
 
 	foreach_array_item(pname0, name->rinstall_if) {
 		name0 = *pname0;
-		if (!name0->ss.seen)
-			continue;
-		if (name0->ss.no_iif)
-			continue;
+		if (!name0->ss.seen) continue;
+		if (name0->ss.no_iif) continue;
 		name0->ss.reevaluate_iif = 1;
 		queue_dirty(ss, name0);
 	}
 }
 
+static void reevaluate_reverse_installif_pkg(struct apk_solver_state *ss, struct apk_package *pkg)
+{
+	struct apk_dependency *d;
+	reevaluate_reverse_installif(ss, pkg->name);
+	foreach_array_item(d, pkg->provides)
+		reevaluate_reverse_installif(ss, d->name);
+}
+
 static void disqualify_package(struct apk_solver_state *ss, struct apk_package *pkg, const char *reason)
 {
 	struct apk_dependency *p;
@@ -149,7 +155,7 @@ static void disqualify_package(struct apk_solver_state *ss, struct apk_package *
 	reevaluate_reverse_deps(ss, pkg->name);
 	foreach_array_item(p, pkg->provides)
 		reevaluate_reverse_deps(ss, p->name);
-	reevaluate_reverse_installif(ss, pkg->name);
+	reevaluate_reverse_installif_pkg(ss, pkg);
 }
 
 static int dependency_satisfiable(struct apk_solver_state *ss, struct apk_dependency *dep)
@@ -240,8 +246,16 @@ static void discover_name(struct apk_solver_state *ss, struct apk_name *name)
 		dbg_printf("discover %s: max_dep_chain=%d no_iif=%d\n",
 			name->name, name->ss.max_dep_chain, name->ss.no_iif);
 	}
-	foreach_array_item(pname0, name->rinstall_if)
-		discover_name(ss, *pname0);
+	foreach_array_item(p, name->providers) {
+		struct apk_package *pkg = p->pkg;
+		foreach_array_item(pname0, pkg->name->rinstall_if)
+			discover_name(ss, *pname0);
+		foreach_array_item(dep, pkg->provides) {
+			if (dep->name->ss.seen) continue;
+			foreach_array_item(pname0, dep->name->rinstall_if)
+				discover_name(ss, *pname0);
+		}
+	}
 }
 
 static void name_requirers_changed(struct apk_solver_state *ss, struct apk_name *name)
@@ -493,17 +507,6 @@ static void reconsider_name(struct apk_solver_state *ss, struct apk_name *name)
 		name->name, name->ss.has_options, name->ss.reverse_deps_done);
 }
 
-static int count_requirers(const struct apk_package *pkg)
-{
-	int cnt = pkg->name->ss.requirers;
-	struct apk_dependency *p;
-
-	foreach_array_item(p, pkg->provides)
-		cnt += p->name->ss.requirers;
-
-	return cnt;
-}
-
 static int compare_providers(struct apk_solver_state *ss,
 			     struct apk_provider *pA, struct apk_provider *pB)
 {
@@ -597,13 +600,6 @@ static int compare_providers(struct apk_solver_state *ss,
 			return r;
 		}
 
-		/* Prefer highest requirer count. */
-		r = count_requirers(pkgA) - count_requirers(pkgB);
-		if (r) {
-			dbg_printf("    prefer highest requirer count\n");
-			return r;
-		}
-
 		/* Prefer installed */
 		if (!(solver_flags & APK_SOLVERF_UPGRADE)) {
 			r = (pkgA->ipkg != NULL) - (pkgB->ipkg != NULL);
@@ -701,7 +697,10 @@ static void assign_name(struct apk_solver_state *ss, struct apk_name *name, stru
 		}
 	}
 	reevaluate_reverse_deps(ss, name);
-	reevaluate_reverse_installif(ss, name);
+	if (p.pkg)
+		reevaluate_reverse_installif_pkg(ss, p.pkg);
+	else
+		reevaluate_reverse_installif(ss, name);
 }
 
 static void select_package(struct apk_solver_state *ss, struct apk_name *name)
@@ -732,8 +731,8 @@ static void select_package(struct apk_solver_state *ss, struct apk_name *name)
 			if (p->version == &apk_atom_null &&
 			    p->pkg->name->auto_select_virtual == 0 &&
 			    p->pkg->name->ss.requirers == 0 &&
-			    (p->pkg->provider_priority == 0 && name->providers->num > 1)) {
-				dbg_printf("    ignore: virtual package without provider_priority with >1 provider\n");
+			    p->pkg->provider_priority == 0) {
+				dbg_printf("    ignore: virtual package without provider_priority\n");
 				continue;
 			}
 			if (compare_providers(ss, p, &chosen) > 0) {
@@ -840,10 +839,8 @@ static void cset_check_install_by_iif(struct apk_solver_state *ss, struct apk_na
 
 	foreach_array_item(dep0, pkg->install_if) {
 		struct apk_name *name0 = dep0->name;
-		if (!name0->ss.in_changeset)
-			return;
-		if (!apk_dep_is_provided(dep0, &name0->ss.chosen))
-			return;
+		if (!name0->ss.in_changeset) return;
+		if (!apk_dep_is_provided(dep0, &name0->ss.chosen)) return;
 	}
 	cset_gen_name_change(ss, name);
 }
@@ -865,6 +862,19 @@ static void cset_check_removal_by_iif(struct apk_solver_state *ss, struct apk_na
 	}
 }
 
+static void cset_check_by_reverse_iif(struct apk_solver_state *ss, struct apk_package *pkg, void (*cb)(struct apk_solver_state *ss, struct apk_name *))
+{
+	struct apk_name **pname;
+	struct apk_dependency *d;
+
+	if (!pkg) return;
+	foreach_array_item(pname, pkg->name->rinstall_if)
+		cb(ss, *pname);
+	foreach_array_item(d, pkg->provides)
+		foreach_array_item(pname, d->name->rinstall_if)
+			cb(ss, *pname);
+}
+
 static void cset_gen_name_remove_orphan(struct apk_solver_state *ss, struct apk_name *name)
 {
 	struct apk_package *pkg = name->ss.chosen.pkg;
@@ -878,7 +888,6 @@ static void cset_gen_name_remove_orphan(struct apk_solver_state *ss, struct apk_
 
 static void cset_gen_name_change(struct apk_solver_state *ss, struct apk_name *name)
 {
-	struct apk_name **pname;
 	struct apk_package *pkg, *opkg;
 	struct apk_dependency *d;
 
@@ -895,10 +904,7 @@ static void cset_gen_name_change(struct apk_solver_state *ss, struct apk_name *n
 		cset_gen_name_remove_orphan(ss, d->name);
 
 	opkg = pkg->name->ss.installed_pkg;
-	if (opkg) {
-		foreach_array_item(pname, opkg->name->rinstall_if)
-			cset_check_removal_by_iif(ss, *pname);
-	}
+	cset_check_by_reverse_iif(ss, opkg, cset_check_removal_by_iif);
 
 	foreach_array_item(d, pkg->depends)
 		cset_gen_dep(ss, pkg, d);
@@ -906,8 +912,7 @@ static void cset_gen_name_change(struct apk_solver_state *ss, struct apk_name *n
 	dbg_printf("Selecting: "PKG_VER_FMT"%s\n", PKG_VER_PRINTF(pkg), pkg->ss.pkg_selectable ? "" : " [NOT SELECTABLE]");
 	record_change(ss, opkg, pkg);
 
-	foreach_array_item(pname, pkg->name->rinstall_if)
-		cset_check_install_by_iif(ss, *pname);
+	cset_check_by_reverse_iif(ss, pkg, cset_check_install_by_iif);
 
 	cset_track_deps_added(pkg);
 	if (opkg)
@@ -921,7 +926,7 @@ static void cset_gen_name_remove0(struct apk_package *pkg0, struct apk_dependenc
 
 static void cset_gen_name_remove(struct apk_solver_state *ss, struct apk_package *pkg)
 {
-	struct apk_name *name = pkg->name, **pname;
+	struct apk_name *name = pkg->name;
 
 	if (pkg->ss.in_changeset ||
 	    (name->ss.chosen.pkg != NULL &&
@@ -931,8 +936,8 @@ static void cset_gen_name_remove(struct apk_solver_state *ss, struct apk_package
 	name->ss.in_changeset = 1;
 	pkg->ss.in_changeset = 1;
 	apk_pkg_foreach_reverse_dependency(pkg, APK_FOREACH_INSTALLED|APK_DEP_SATISFIES, cset_gen_name_remove0, ss);
-	foreach_array_item(pname, pkg->name->rinstall_if)
-		cset_check_removal_by_iif(ss, *pname);
+	cset_check_by_reverse_iif(ss, pkg, cset_check_removal_by_iif);
+
 	record_change(ss, pkg, NULL);
 	cset_track_deps_removed(ss, pkg);
 }
@@ -1026,13 +1031,16 @@ static int compare_name_dequeue(const struct apk_name *a, const struct apk_name
 {
 	int r;
 
-	r = (!!a->ss.requirers) - (!!b->ss.requirers);
+	r = !!(a->solver_flags_set) - !!(b->solver_flags_set);
 	if (r) return -r;
 
 	r = (int)a->priority - (int)b->priority;
 	if (r) return r;
 
 	r = a->ss.max_dep_chain - b->ss.max_dep_chain;
+	if (r) return -r;
+
+	r = (!!a->ss.requirers) - (!!b->ss.requirers);
 	return -r;
 }
 
